/*
 * MaxClassPol.h
 *
 *  Created on: 26 feb 2012
 *      Author: Wolfcastle
 *
 *  Description
 *		This library implements the independent t-test
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

extern int indTTest(double**, int, int, char**, char*, char*);
