package Algorithm::TrunkClassifier::FeatureSelection;

use warnings;
use strict;

require XSLoader;

our $VERSION = 'v1.0.1';

XSLoader::load('Algorithm::TrunkClassifier::FeatureSelection', $VERSION);

return 1;
